# TUCOPY Tucows mirroring software
# Copyright (C) 2001 Tucows Inc.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. 
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. 
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


#
# vi:set tabstop=4:
# $Id: Fetch.pm,v 1.36 2001/10/23 20:01:23 lim Exp $
#
# Authors:
#       Bojan Jovanovic <bjovanovic@tucows.com>
#       Leonid Igolnik <lim@tucows.com>
#

package Tucows::Mirror::Client::Fetch;
use Tucows::Mirror::Config;
use Tucows::Mirror::Log;
use File::Basename;
use File::Path;
use Tucows::Mirror::Client::Fetch::Error_codes qw( get_error_type );
use Tucows::Mirror::Client::Uncompress qw(uncompress_file);

require Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(fetch_file fetch_record);

use strict;

my ($_config, $_delivery_mechanism, $_delivery_dr);
my $_init = 0;

#Default delay between attempts, used if package has no defenition.
my $_delay = 30; 

sub _init {
	logit('debug', 7, 'D1 Initalizing');
	$_config = Tucows::Mirror::Config->new() || die "\n";
	$_delivery_mechanism = $_config->delivery_mechanism;
	$_delivery_dr = 'Tucows::Mirror::Client::Fetch::' . $_delivery_mechanism .
					'::fetch';

	# Version prior to 5.005 does not like exec/exit combination
	my $warn_flag = $^W;
	$^W = 0 if ($] < 5.005);

	eval("require Tucows::Mirror::Client::Fetch::$_delivery_mechanism");

	# return the flag to original state
	$^W = $warn_flag;

	if ($@) {
		die "Unable to initialize driver for $_delivery_mechanism: $@";
	} 

	logit('debug', 7, 'D2 Initalization completed');
	$_init = 1;
}


#Parameters for _fetch_file: in HASHREF:
#	URL: url to fetch
#	Target: Final file location
#	User: Username to use for aythentication
#	Pass: Passowrd to use
# Note: _fetch_file is internal function, and should not be called from the
#		code outside this module
sub _fetch_file {
	my @args = @_;
	
	_init() unless $_init;
	
	logit('debug', 8, 'D3 Calling delivery driver \'%s\'', $_delivery_dr);
	no strict 'refs';
	return &$_delivery_dr(@args);
}

sub build_url {
	my ($pkg, $filename, $compressed) = @_;

	# url encode the filename (except for /)
	$filename =~ s#([^a-zA-Z0-9_.-/])#uc sprintf("%%%02x", ord($1))#eg;
	# remove double /
	$filename =~ s#/+#/#g;

	my $url = $pkg->{package_url} . $filename;
	my $params='';

	$params .= 'speed=' . $pkg->{download_speed} . '&'
					if $pkg->{download_speed};
	$params .= 'compressed=' . $compressed . '&' if $compressed;
	$params .= 'sessid=' . $$ . '&';
	$params .= 'cversion='.$main::VERSION.'&';

	$url .= '?' . $params;
	logit('debug', 9, 'D4 request URL is %s', $url);
	return $url;
}


# Usage
# fetch_file({filename => 'index.tx',pkg => $package, 
# Target => '/tmp/1', compressed => 0}, size => 10);

sub fetch_file {
	my ($params) = @_;
	my $pkg = $params->{pkg} || return 0;

	logit('debug', 9, 'D5 Processing request for %s', $params->{filename});

	my $url = build_url($pkg, $params->{filename}, $params->{compressed});
	
	my $tries = 0;
	my $status;

	while ( ( not defined $status or not $status->{success} ) and
		$tries < $pkg->{ max_number_of_tries } and 
		( not defined $status or $status->{retry} ) ) {

		# Get the file
		# Format of $status is defined in 
		# Tucows::Mirror::Client::Fetch::Error_codes

		$status = _fetch_file({	Target => $params->{ Target },
									URL => $url,
									User => $pkg->{ username },
									Pass => $pkg->{ password }
								});
		$tries++;

		if ( $status->{ success } ) {
			if (defined $params->{size} && $params->{size} > 0) {
				my $fsize = (stat($params->{Target}))[7];
				$status = get_error_type( 
					'file_size_mismatch', $params->{Target}, $fsize, $params->{size} )
					unless ($fsize == $params->{size});
			} 
		} else {
			my $delay = $pkg->{delay_before_retry} || $_delay;
			sleep($delay*$tries);
		}

	}
	return $status;
}

sub fetch_record {
	my ($config, $params, $pkg, $rec, $base_dir, $rename) = @_;
	
	my $use_compression = $rec->is_compressed && $config->use_compression;
	my $target = $pkg->package_root . '/' . $base_dir . '/' . $rec->file_name;
	my $dir = dirname($target);

	unless (-d $dir) {
		logit('debug', 6, 'D6 Creating dir %s', $dir);
		mkpath($dir, 0, oct($config->directory_permissions));
	}

	my $start_time = time;

	my $target_size = 0;
	unless ($params->{ignoresize}) {
		$target_size = $use_compression ?
							$rec->file_size_compressed() : $rec->file_size
	}

	my $result = fetch_file({
            filename => $base_dir.'/'.$rec->file_name,
            pkg => $pkg,
            Target => $target . '.tmp' . ($use_compression ? '.gz' : ''),
            compressed => $use_compression,
			size => $target_size,
        }
	);
	my $elapsed_time = time - $start_time;

	if (!$result->{success}) {
		logit('error', 2, 'E2 Failed to fetch file: %s, Reason: %s',
				$base_dir.'/'.$rec->file_name, $result->{message});
		unlink($target . '.tmp');
		return 0;
	}
	
	if ($use_compression) {
		uncompress_file($target . '.tmp.gz', $target . '.tmp') || do {
			logit('error', 4, 'E3 Unable to uncompress %s.tmp.gz', $target);
			unlink($target . '.tmp.gz');
			return 0;
		};
		unlink($target . '.tmp.gz') || do {
			logit('error', 5, 'E4 Unable to unlink %s.tmp.gz: %m', $target);
		};
	}

	utime($rec->mtime, $rec->mtime, $target . '.tmp') || do {
		logit('error', 3, 'E5 Failed to set timestamp on %s to %s: %m',
				$target . '.tmp', $rec->mtime);
		unlink($target . '.tmp');
		return 0;
	};

	if ($rename) {
		rename($target . '.tmp', $target) || do {
			logit('error', 3, 'E6 Unable to rename %s to %s: %m',
					$target . '.tmp', $target);
			unlink($target . '.tmp');
			return 0;
		};
	}

	logit('stats', 0, '%d %s "%s" %d %d %d', $$, $pkg->package_name, 
			$target, $rec->file_size,
			$use_compression ? $rec->file_size_compressed : 
								$rec->file_size,
			$elapsed_time);
	
	return 1;
}

1;
