# TUCOPY Tucows mirroring software
# Copyright (C) 2001 Tucows Inc.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. 
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. 
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


#
# vi:set tabstop=4:
# $Id: Wget.pm,v 1.17 2001/05/07 17:10:09 bjovanovic Exp $
#
# This module contains Wget file fetching driver
#
# Authors:
#		Bojan Jovanovic <bjovanovic@tucows.com>
#       Alex Nobert <anobert@tucows.com>
#		Leonid Igolnik <lim@tucows.com>



package Tucows::Mirror::Client::Fetch::Wget;

use strict;
use integer;
use Tucows::Mirror::Config;
use Tucows::Mirror::Client::Fetch::Error_codes qw( ok_code get_error_type );

# Has init been called?
my $_init = 0;

# following variables will be initialized in _init
my $_config;
my $_wget_location;
my $_tmp_dir;
my $_wget_error_log;

# constants

# driver name
my $_driver = 'Wget';

# Error codes returned from the fork
my $_exec_failed_code = 201; #Should not colide with any of the wget error codes

# Null device
my $_null = '/dev/null'; 


# _init function - called only once
sub _init {
	$_config = Tucows::Mirror::Config->new();
	$_wget_location = $_config->delivery_wget_location;
	$_tmp_dir = $_config->tmp_dir();
	$_wget_error_log = "$_tmp_dir/.wget_log";
	$_init = 1;
	return 1;
}

sub fetch {
	my ($opts) = shift;

	_init()
		unless ($_init);

	# Prepare authorization header
	my $headers = 'Authorization: Basic '.
		_encode_base64("$opts->{User}:$opts->{Pass}");

	# fork. Child exec's Wget, parent waits
	my $pid = fork;

	return get_error_type('fork_failed', $$)
		unless (defined $pid);

	if ( $pid ) {
		# parent waits on child to complete
		waitpid($pid,0)
			or return get_error_type('wait_on_pid', $pid);
	} else {
		$| = 1;
		close(STDOUT);
		# redirect all output to null
		open(STDOUT, ">$_null") or exit( $_exec_failed_code );
		close(STDERR);
		# redirect all output to null
		open(STDERR, ">>$_wget_error_log") or exit( $_exec_failed_code );
		# the exec itself
		exec( $_wget_location, $opts->{URL}, '-nv', "--header=$headers", 
			"--output-file=$_wget_error_log", '-O', $opts->{Target});

		# code should never get here. If it did, that means exec failed
		exit( $_exec_failed_code );
	}

	my $status;

	# if exit code was not 0 that means there was some kind of error
	if ($?) {

		# is it the exec_failed error we caused to report a problem?
		return get_error_type( 'exec_failed', $_wget_location )
			if ( $? >> 8 == $_exec_failed_code or $? & 127 or $? &128 );

		# let's open log file and see what was the problem
		open(ERR_LOG, $_wget_error_log) 
			or return get_error_type('log_open_error', $_wget_error_log);

		# get the lines in an array
		my @lines = <ERR_LOG>;

		close(ERR_LOG);

		# only the last line is relevant
		my $last_line = $lines[$#lines];

		# get the HTTP error code
		return get_error_type('log_parse_error', $_driver, $_wget_error_log, "\n@lines")
			unless ( $last_line =~ /\d\d\:\d\d:\d\d\s+ERROR\s+(\d+):/ ); 

		# return exit status based on the HTTP error code
		$status = get_error_type( $1, $last_line );

	} else {

		# We assume if no error code was returned from wget, return code was OK
		$status = get_error_type( ok_code() );
	}

	# unlink the log file
	unlink $_wget_error_log;

	return $status;

}

# Stripped perl version of encode base64 function
sub _encode_base64 ($)
{
	my $res = "";

	pos($_[0]) = 0;						# ensure start at the beginning

	while ($_[0] =~ /(.{1,45})/gs) {
		$res .= substr(pack('u', $1), 1);
		chop($res);
	}

	$res =~ tr|` -_|AA-Za-z0-9+/|;		

	# fix padding at the end
	my $padding = (3 - length($_[0]) % 3) % 3;
	$res =~ s/.{$padding}$/'=' x $padding/e if $padding;

	return $res;
}

1;
