# TUCOPY Tucows mirroring software
# Copyright (C) 2001 Tucows Inc.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. 
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. 
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


#
# vi:set tabstop=4:
# $Id: Listing.pm,v 1.32 2001/05/10 20:36:10 bjovanovic Exp $
# 
# Authors:
#	   Bojan Jovanovic <bjovanovic@tucows.com>
#	   Leonid Igolnik <lim@tucows.com>
#

package Tucows::Mirror::Client::Listing;

use Tucows::Mirror::Client::Fetch qw( fetch_record );
use Tucows::Mirror::Client::Lock qw( touch_lock_file );
use Tucows::Mirror::Log;
use File::Basename;
use File::Path;

require Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw( process_listing );


use strict;

sub process_listing {
	my ($package, $listing_record, $config, $params, $delete_list, $action) =
			@_;
	my $status = 1;

	# base directory for the listing relative to the package root
	my $base_dir = dirname($listing_record->file_name());

	# create the directory if it does not exist
	my $dir = $package->package_root().'/'.$base_dir;
	unless (-d $dir) {
		logit('debug', 6, 'D22 Creating dir %s', $dir);
		mkpath($dir, 0, oct($config->directory_permissions));
	}



	my $old_listing = Tucows::Mirror::Listing->new();
	my $get_pattern = $package->get_pattern;

	$old_listing->build( {
		root_dir => $package->package_root().'/'.$base_dir,
		separate_index_files => 1,
		recursive => 0,
		build_dir => '',
		file_pattern => '*',}
	) || do {
		logit('error', 2, 'E40 Unable to build local listing for %s/%s', 
				$package->package_root, $base_dir);
	};


	my $new_listing = Tucows::Mirror::Listing->new();
	if ($action eq 'update') {	
		fetch_record($config, $params, $package, $listing_record, '', 1) || do {
			logit('error', 2, 'E41 Unable to fetch remote listing %s',
					$listing_record->file_name);
			return 0;
		};

		$new_listing->load( $package->package_root().'/'.
			 $listing_record->file_name()) || do {
				logit('error', 2, 'E42 Failed to load downloaded listing: %s',
						$package->package_root() . '/' .
						$listing_record->file_name() );
			return 0;
		};
		if (not $package->keep_listing_files()) {
			unlink( $package->package_root().'/'. $listing_record->file_name() )
				or logit('error', 4, 'E55 Unable to remove listing file: %s, reason: %m',
					$package->package_root().'/'. $listing_record->file_name());
		}
	}

	my ($diff_listing, $change_count) = 
		$new_listing->compare_lists( $old_listing );

	logit('debug', 6, 'D23 Detected %d changes in the %s listing file', 
		$change_count, $package->package_root().'/'
		.$listing_record->file_name() );

	$diff_listing->rewind();

	my $lock_filename = $config->tmp_dir.'/.mirror.'.$package->package_name();

	while (my $record = $diff_listing->next()) {

		touch_lock_file( $lock_filename ) ||
			logit('error', 4, 'E56 Failed to refresh timestamp on the lock file: %m',
			$lock_filename);
			

		if ( $record->action() eq 'update') {
			update_record($package, $record, $config, $params, $base_dir) ||
				do {
				$status = 0;
			};
		} elsif ( $record->action() eq 'delete') {
			delete_record($package, $record, $config, $params, $base_dir,
							$delete_list) || do {
				$status = 0;
			};
		}
	}
	return $status;
}

sub delete_record {
	my ($package, $record, $config, $params, $base_dir, $delete_list) = @_;

	return 1 if $params->sync;
	
	# Don't delete catalog temp file as it has to be renamed upon completion
	return 1 if ($record->file_name() eq '.mirroridx.idx.tmp');

	# anything else that appeared in the local listing should be
	# scheduled for deletion
	if (defined $package->delete_pattern && $package->delete_pattern) {
		my $delete_pattern = $package->delete_pattern;
		if ($record->file_name !~ /$delete_pattern/) {
			logit('debug', 6, 'D25 Will not delete %s, does not match %s', 
					$record->file_name, $package->delete_pattern);
			return 1;
		}
	}

	if (defined $package->delete_ignore_pattern &&
			$package->delete_ignore_pattern) {
		my $delete_ignore_pattern = $package->delete_pattern;
		if ($record->file_name =~ /$delete_ignore_pattern/) {
			logit('debug', 6, 'D26 Will not delete %s, matches %s',
					$record->file_name, $package->delete_ignore_pattern);
			return 1;
		}	
	}

	logit('debug', 9, 'D27 Scheduling for deletion %s', $package->package_root().
			'/'.$base_dir.'/'.$record->file_name() );

	if ($params->t) {
		printf("Delete %s/%s/%s\n", $package->package_root, $base_dir, 
				$record->file_name);
	} else {
		push @$delete_list, {
			filename => $package->package_root().  
						'/'.$base_dir.'/'.$record->file_name(),
			is_directory => $record->is_directory(),
		};
	}
}

sub update_record {
	my ($package, $record, $config, $params, $base_dir) = @_;
	
	return 1 if $record->is_directory;

	if (defined $package->get_pattern && $package->get_pattern) {
		my $get_pattern = $package->get_pattern;
		if ($record->file_name !~ /$get_pattern/) {
			logit('debug', 6, 'D28 File %s does not match get_pattern %s, skipping',
					$record->file_name, $package->get_pattern);
			return 1;
		}
	}

	if (defined $package->get_ignore_pattern && $package->get_ignore_pattern) {
		my $get_ignore_pattern = $package->get_ignore_pattern;
		if ($record->file_name =~ /$get_ignore_pattern/) {
			logit('debug', 6, 'D29 File %s matches get_ignore_pattern %s, skipping',
					$record->file_name, $package->get_ignore_pattern);
			return 1;
		}
	}

	if ($params->t) {
		printf("Update %s/%s/%s\n", $package->package_root, $base_dir,
					$record->file_name);
		return 1;
	}

	
	if (!$params->sync) {
		logit( 'debug', 9, '30 Updating %s', $package->package_root().
				'/'.$base_dir.'/'.$record->file_name() );

		fetch_record($config, $params, $package, $record, $base_dir, 1) || do {
			logit('error', 2, 'E44 Was unable to update file %s',
					$record->file_name());
			return 0;
		};
	} else {
		timestamp_record($config, $params, $package, $record, $base_dir) || do {
			logit('error', 2, 'E45 Unable to timestamp listing %s/%s/%s',
					$package->package_root, $base_dir, $record->file_name);
			return 0;
		};
	}
}

sub timestamp_record {
	my ($config, $params, $package, $rec, $base_dir) = @_;

	# if file is not there there is nothing to timestamp (this function is called only
	# from -sync)..
	return 1 unless -e $package->package_root . '/' . $base_dir . '/' .
							$rec->file_name;

	logit('debug', 7, 'D31 Updating timestamp on  %s/%s/%s to %s',
			$package->package_root, $base_dir, $rec->file_name, $rec->mtime);

	utime($rec->mtime, $rec->mtime, $package->package_root . '/' .
			$base_dir . '/' . $rec->file_name) || do {
		logit('error', 3, 'E46 Failed to set timestamp on %s/%s/%s to %s: %m',
		$package->package_root, $base_dir, $rec->file_name, $rec->mtime);
			return 0;
	};

	return 1;
}

1;
