# TUCOPY Tucows mirroring software
# Copyright (C) 2001 Tucows Inc.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. 
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. 
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


# 
# vi:set tabstop=4:
# $Id: Package.pm,v 1.45 2001/11/06 16:53:08 bjovanovic Exp $
#
# Client package processing routines
#
# Authors:
#	Leonid Igolnik <lim@tucows.com>
#	Bojan Jovanovic <bjovanovic@tucows.com>
#
package Tucows::Mirror::Client::Package;

@ISA = (Tucows::Mirror::Package);

use strict;
use Tucows::Mirror::Package;
use Tucows::Mirror::Log;
use Tucows::Mirror::Listing;
use Tucows::Mirror::Client::Listing qw( process_listing );
use Tucows::Mirror::Client::Lock qw(lock_file unlock_file);
use Tucows::Mirror::Client::Fetch qw(fetch_file fetch_record);
use File::Path;
use IO::File;

sub new {
	my $module_name = shift;
	my $file_name = shift;
	my $pkg = $module_name->SUPER::new($file_name);
	if ($pkg) {
		$pkg->{_delete_ok} = [];
	} else {
		return undef;
	}
	return $pkg;
}

sub get_packages {
	my $module_name = shift;
	my ($package_names, $config) = @_;

	# get the default package directory
	my $package_directory = $config->package_dir();

	my @package_list = @$package_names;
	
	# If list of packages is not supplied, try to load it from the default
	# location
	my $package_file_names = (@$package_names) ? $package_names :
		get_default_package_filenames($package_directory);

	my @packages;

	foreach my $package_file_name (@$package_file_names) {
		push @packages, $module_name->new( $package_file_name );
	}

	return \@packages;
}

# gets the list of packages from the supplied directory
sub get_default_package_filenames {
	my $package_directory = shift;

	# make sure that directory has the last /
	$package_directory .= '/'
		unless ($package_directory =~ /\/$/);

	opendir(INP_DIR, $package_directory) or 
		do {
			logit('error', 2, 'E12 Unable to open package directory %s',
					$package_directory);
			return [];
		};

	my @files;
	while (defined(my $entry = readdir(INP_DIR)))
		{
		my $fullname = $package_directory.$entry;
		# accept only files that are ASCII files, non-zero size, and readable
		push @files, $package_directory.$entry
			if ( $entry =~ /\.pkg$/i and 
			-T $package_directory.$entry and 
			-s $package_directory.$entry and
			-r $package_directory.$entry );
		}

	closedir(INP_DIR);

	return \@files;
}

sub process {
	my ($self, $config, $params) = @_;
	my $status = 1;

	# $deletions keeps all files/directories to be deleted
	# part of safety switch - to prevent deletion of the whole tree
	my $deletions = [];

	logit('debug', 4, 'D7 Processing %s package', $self->package_name);

	logit('debug', 5, 'D8 Locking %s package', $self->package_name);
	my $lock_filename = $config->tmp_dir . '/.mirror.' . $self->package_name;
	lock_file($lock_filename, $self->lock_timeout) || do {
		logit('error', 3, 'E13 Unable to lock the package');
		return 0;
	};
	
	unless (-d $self->package_root) {
		mkdir($self->package_root, oct($config->directory_permissions)) || do {
			logit('error', 2, 'E14 Unable to mkdir %s: %m', $self->package_root);
			unlock_file($lock_filename);
			return 0;
		};
	}

	unless ($params->t() or $params->sync()) {
		logit('debug', 6, 'D9 Fetching .mirror_timestamp_a file');
		my $result = fetch_file({
				filename => 'get_time_stamp',
				pkg => $self,
				Target => $self->package_root . '/.mirror_timestamp_a',
				compressed => 0,
				'rename' => 0,
			}
		);
		if (!$result->{success}) {
			logit('error', 2, 'E15 Failed to fetch .mirror_timestamp_a file: %s',
					$result->{message});
		}
	}

	logit('debug', 6, 'D10 Fetching catalog file');
	my $result = fetch_file({
			filename => '.mirroridx.idx',
			pkg => $self,
			Target => $self->package_root . '/.mirroridx.idx.tmp',
			compressed => 0,
			'rename' => 0,
		}
	);
	if (!$result->{success}) {
		logit('error', 2, 'E16 Failed to fetch catalog file: %s',
				$result->{message});
		unlock_file($lock_filename);
		return 0;
	}

	$self->init_delete_ok();

	my $local_listing = Tucows::Mirror::Listing->new();
	if (-s $self->package_root . '/.mirroridx.idx' &&
			!$params->force && !$params->sync) {
		$local_listing->load($self->package_root . '/.mirroridx.idx') || do {
			logit('error', 3,
				'E17 Unable to read LOCAL catalog file from %s/%s',
				$self->package_root, '.mirroridx.idx');
			return 0;
		};
	}

	my $remote_listing = Tucows::Mirror::Listing->new();
	$remote_listing->load($self->package_root . '/.mirroridx.idx.tmp') || do {
		logit('error', 3, 'E18 Unable to read REMOTE catalog file from %s/%s',
			$self->package_root, '.mirroridx.idx.tmp');
		return 0;
	};

	my ($changes, $num_changes) =
			$remote_listing->compare_lists($local_listing);
	logit('debug', 4, 'D11 Detected %d changes in the catalog file', $num_changes);
	while (my $record = $changes->next) {
		if ($record->action eq 'delete') {
			logit('debug', 5, 'D12 %s was removed from catalog file',
					$record->file_name);
			process_listing($self, $record, $config, $params, $deletions,
								'delete') || do {
				logit('error', 2, 'E19 Deletion of listing %s was not successful',
						$record->file_name);
				$status = 0;
			};
		} elsif ($record->action eq 'update') {
			logit('debug', 5, 'D13 %s was updated, running update',
					$record->file_name);
			process_listing($self, $record, $config, $params, $deletions,
								'update') 
				or do {
				logit('error', 2, 'E20 Update of listing %s was not successful', 
						$record->file_name);
				$status = 0;
			};
		} else {
			logit('error', 2, 'E21 Unknown action %s', $record->action);
		}
	}

	logit('debug', 5, 'D14 Processing deletions');
	$self->delete_scheduled($deletions, $self->deletion_limit()) or
		$status = 0;

	# get completed timestamp
	unless ($params->t() or $params->sync()) {
		logit('debug', 6, 'D15 Fetching .mirror_timestamp_z file');
		my $result = fetch_file({
				filename => 'get_time_stamp',
				pkg => $self,
				Target => $self->package_root . '/.mirror_timestamp_z'.$status,
				compressed => 0,
				'rename' => 0,
			}   
		);	  
		if (!$result->{success}) {
			logit('error', 2, 'Failed to fetch .mirror_timestamp_z file: %s',
					$result->{message});
		}
	}

	if ($status && !$params->t) {
		logit('debug', 5, 'D16 Replacing old catalog file');
		rename($self->package_root . '/.mirroridx.idx.tmp',
				$self->package_root . '/.mirroridx.idx') || do {
			logit('error', 3, 'E23 Unable to rename %s/%s to %s/%s: %m',
					$self->package_root, '.mirroridx.idx.tmp',
					$self->package_root, '.mirroridx.idx');
			unlock_file($lock_filename);
			return 0;
		};
	} else {
		logit('error', 3, 'E24 Will not replace catalog file for package %s',
				$self->package_name);
		print STDERR "An error occured during processing of ".$self->package_name.
			" package.\n\tPlease check error log at: ".
			join("; ", @{get_log_filenames('error')})."\n";
	}

	logit('debug', 5, 'D17 Unlocking package');
	unlock_file($lock_filename) || do {
		logit('error', 3, 'E25 Unable to unlock package');
		return 0;
	};

	logit('debug', 4, 'D18 Done processing %s package', $self->package_name);
	return 1;
}

sub delete_scheduled {
	my ($self, $deletions, $deletion_limit) = @_;

	my $status = 1;


	my $count = 0;
	foreach my $deletion (@$deletions) {
		$count++ 
			if (not $self->ignore_delete_limit($deletion->{filename},
					$deletion->{is_directory}));
	}

	if ( $count > $deletion_limit ) {
		logit('error', 2, 'E26 %d deletions attempted. Maximum %d! Skipping all deletions',
			 $count, $deletion_limit);
		return 0;
	} else {

		my $directory_deletions = [];
		# delete files and push direcories to $directory_deletions
		foreach my $deletion (@$deletions) {
			if ($deletion->{is_directory}) {
				$deletion->{filename} =~ s/\/+/\//g;
				push @$directory_deletions, $deletion->{filename};
				next;
			}
			logit('debug', 5, 'D19 Deleting %s',$deletion->{filename});

			unlink $deletion->{filename} or do {
					logit('error', 2, 'E27 Unable to delete %s', 
						$deletion->{filename});
					$status = 0;
				};
		}

		my @sorted_directory_deletions = sort {$b cmp $a} @$directory_deletions;

		foreach my $directory (@sorted_directory_deletions) {
			logit('debug', 5, 'D20 Removing (direcory) %s', $directory);
			rmdir $directory or do {

					#check if directory matches one of the directories listed in
					# delete_unknown_dir_ignore

					my $delete_unknown_dir_ignore = $self->delete_unknown_dir_ignore();
					# remove leading and trailing pipes
					$delete_unknown_dir_ignore =~ s/\|+$//;
					$delete_unknown_dir_ignore =~ s/^\|+//;

					my @directories_to_ignore = split(/\|/, $delete_unknown_dir_ignore);
					my $match_found = '';
					
					$directory =~ s/\/+/\//g;
					my $directory_quoted = quotemeta($directory);

					# directory names to match are stored in pipe separated list
					foreach my $directory_to_match (@directories_to_ignore) {
						if ($directory_to_match =~ /^$directory_quoted/) {
							$match_found = $directory_to_match;
							last;
						}
					}
					if ($match_found) {
						logit('debug', 7, 'D32 Will not remove unknown directory %s'.
							' - matching %s from  %s', $directory, $match_found,
							$delete_unknown_dir_ignore);
					} else {
						logit('debug', 7, 'D33 Removing unknown directory and all '.
							'subdirectories: %s', $directory);
						rmtree($directory, 0, 1);
					}
				};
		}

	}
	return $status;
}

sub init_delete_ok {
	my ($self) = @_;
	logit('debug', 6, 'D21 Fetching .mirror_delete_ok file');
	my $result = fetch_file({
			filename => '.mirror_delete_ok',
			pkg => $self,
			Target => $self->package_root . '/.mirror_delete_ok',
			compressed => 0,
		});

	if (!$result->{success}) {
		logit('error', 2, 'E29 Failed to fetch /.mirror_delete_ok file: %s',
			$result->{message});
		return 0;
	}

	my $fh = IO::File->new($self->package_root . '/.mirror_delete_ok', 'r') || do {	
		logit('error', 3, 'E30 Failed to open %s/%s: %m', 
				$self->package_root, '.mirror_delete_ok');
		return 0;
	};

	while (defined(my $dir = <$fh>)) {
		chomp $dir;
		my $local_dir = $self->package_root . '/' . $dir;
		$local_dir =~ s#/+#/#g;
		push @{$self->{_delete_ok}}, $local_dir;
	}

	$fh->close;	

	return 1;
}

sub ignore_delete_limit {
	my ($self, $filename) = @_;
	$filename =~ s#/+#/#g;

	foreach my $ignore (@{$self->{_delete_ok}}) {
		return 1 if $filename =~ /^$ignore/;
	}
	return 0;
}

1;
