# TUCOPY Tucows mirroring software
# Copyright (C) 2001 Tucows Inc.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. 
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. 
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


# 
# vi:set tabstop=4:
# $Id: Listing.pm,v 1.57 2001/05/04 21:17:50 bjovanovic Exp $
#
# Listing files implementation
#
# Authors: 
#	Leonid Igolnik <lim@tucows.com>
#	Bojan Jovanovic <bjovanovic@tucows.com>
#	Alex Nobert <anobert@tucows.com>
#
package Tucows::Mirror::Listing;

use IO::File;
#use IO::Dir;
use Tucows::Mirror::Listing::Record;
use Tucows::Mirror::Log;
use strict;

my $_tmp_ext = '.tmp';

#Creates new listing object
sub new {
	my ($pkg, $params) = @_;
	my $listing = {
		_data => [],
		_lookup => {},
		_index => 0,
		_filename => undef,
		_server => $params->{server},
	};
	bless $listing, $pkg;
}

#Loads listing from a file
#Parameters:	$filname - listing filename
#Returns:		0/1
sub load {
	my $self = shift;
	my $filename = shift;
	my $fh = IO::File->new($filename, 'r') || return 0;
	$self->{_filename} = $filename;

	my $last_rec_found = 0;
	while (<$fh>) {
		## Required if loading a file that has less fields than we support!
		chomp;

		#Check if current line is a last record
		if (/^-1\|-1\|/) {
			$last_rec_found = 1;
			last;
		}

		#Add record to the listing
		$self->add_rec(Tucows::Mirror::Listing::Record->new({record => $_ }));
	}	

	$fh->close;

	logit('error', 3, 'E54 Incomplete listing file (no last record): %s',
			$filename) unless $last_rec_found;

	return $last_rec_found;	
}

sub build {
	my $self = shift;
	my $params = shift;

	my $root_dir = $params->{root_dir};
	my $build_dir = $params->{build_dir} || '';
	my $compressed_root_dir = $params->{compressed_root_dir} || '';
	my $file_pattern = $params->{file_pattern} || '*';
	my $save_file = $params->{save_file} || '';

	# Make a copy of the pattern match and make it useable in //;
	my $match = $file_pattern;
	$match =~ s/\./\\\./g;
	$match =~ s/\*/\.\*/g;
	$match =~ s/\?/\.\?/g;

	opendir(LS, $root_dir . '/'. $build_dir) || do {
				logit('error', 2, 'E47 Unable to open dir %s: %m', 
						$root_dir . '/'. $build_dir);
				return 0;
	};
	my(@files) = readdir(LS);
	closedir(LS);

	foreach my $file_name (@files) {

		# files with '|' in filename should be skipped
		# (not supported - | is delimiter in catalog/listing files)
		next
			if ($file_name =~ /\|/);

		if ($file_pattern eq '*') {
			# Skip special files if called with wildcard 
			next if ($file_name =~ /^\.\.?$/ or $file_name =~ /^\.mirror/  );
		} else {
			# skip . and .. 
			next if ($file_name =~ /^\.\.?$/);
		}

		next if ($file_name !~ /^$match$/ &&
					! -d "$root_dir/$build_dir/$file_name");

		my ($rec_root_dir, $rec_compressed_root_dir, $rec_file_name);

		if ($params->{separate_index_files}) { 
			$rec_root_dir = $root_dir . '/' . $build_dir;
			$rec_compressed_root_dir = $compressed_root_dir . '/' . $build_dir;
			$rec_file_name = $file_name;
		} else {
			$rec_root_dir = $root_dir;
			$rec_compressed_root_dir = $root_dir;
			$rec_file_name = $build_dir . '/' . $file_name;
		}
		my $rec = Tucows::Mirror::Listing::Record->new({
						root_dir => $rec_root_dir,
						compressed_root_dir => $rec_compressed_root_dir,
						file_name => $rec_file_name,
						server => $self->{_server},
					});

		if (-d $root_dir . '/' . $build_dir . '/' . $file_name and 
			$params->{recursive}) {
			if ($params->{separate_index_files}) {
				## Create a new listing object for each directory and build it
				my ($listing) = Tucows::Mirror::Listing->new({
									server => $self->{_server},
								});
				$listing->build({
						root_dir => $root_dir,
						build_dir => $build_dir . '/' . $file_name,
						compressed_root_dir => $compressed_root_dir,
						file_pattern => $file_pattern,
						save_file => $save_file,
						no_list_directories => $params->{no_list_directories},
						separate_index_files => $params->{separate_index_files},
						recursive => $params->{recursive},
					}) || do {
						logit('error', 3,
							'E48 Unable to build listing for %s/%s/%s: %m',
							$root_dir, $build_dir, $file_name);
						return 0;
				};
			} else {
				$self->build({
						root_dir => $root_dir,
						build_dir => $build_dir . '/' . $file_name,
						compressed_root_dir => $compressed_root_dir,
						file_pattern => $file_pattern,
						save_file => $save_file,
						no_list_directories => $params->{no_list_directories},
						recursive => $params->{recursive},
					}) || do {
						logit('error', 3,
							'Unable to build listing for %s/%s/%s: %m',
							$root_dir, $build_dir, $file_name);
						return 0;		
				};
			}
		}

		$self->add_rec($rec)
			unless (-d $root_dir . '/' . $build_dir . '/' . $file_name and
			$params->{no_list_directories});

	}

	## If save_file is defined, save to a file
	if ($save_file) {
		$self->save("$root_dir/$build_dir/$save_file") || do {
			logit('error', 3, 'E49 Unable to save a listing file %s/%s/%s: %m',
					$root_dir, $build_dir, $save_file);
			return 0;
		}
	}

	return 1;
}

## $list_obj->compare_lists($old_list);
sub compare_lists {
	my $self = shift;
	my $old_listing = shift;

	my ($diff_listing) = Tucows::Mirror::Listing->new();

	# count the number of changes
	my $changes = 0;

	$self->rewind;

	while (defined (my $new_record = $self->next)) {
		my $old_record = $old_listing->find($new_record->file_name);

		# If record is completely new, or if it's out of date
		if (not defined $old_record or 
			not $new_record->compare_record($old_record, $self->{_server})) {
			$new_record->action('update');
			$diff_listing->add_rec($new_record);
			# Increase the number of changes if new record or if not directory
			# (ignore changes in directory timestamp)
			$changes++
				if (not defined $old_record or not $old_record->is_directory());
		}
	}

	$old_listing->rewind;

	while (defined (my $old_record = $old_listing->next)) {
		my $new_record = $self->find($old_record->file_name);

		# if old record has been deleted
		if (! defined $new_record ) {
			# Here we assume that old_listing will NOT be reused!
			$old_record->action('delete');
			$diff_listing->add_rec($old_record);
			$changes++;
		}
	}

	return($diff_listing, $changes);
}

sub add_rec {
	my $self = shift;
	my $rec = shift;
	push @{$self->{_data}}, $rec;
	$self->{_lookup}->{$rec->file_name()} = $#{$self->{_data}}; 
	return $rec;
}

sub save {
	my $self = shift;
	my $filename = shift || $self->{_filename} || return 0;

	if (-f $filename) {
		## Load existing listing file
		my $list_old = Tucows::Mirror::Listing->new();
		my $list_changed = Tucows::Mirror::Listing->new();
		$list_old->load($filename) || do {
			logit('error', 3, 'E50 Unable to read listing from %s', $filename);
			return 0;
		};

		my $changes = 0;
		my $changes_local = 0;
		## Compare lists, return if no changes have been made
		($list_changed,$changes) = $self->compare_lists($list_old);
		($list_changed,$changes_local) = $list_old->compare_lists($self);
		if ( ($changes+$changes_local) < 1) { return 1; }
	}

	## Open save file for writing
	my $fh = IO::File->new($filename . $_tmp_ext, 'w') || do {
		logit('error', 3, 'E51 Unable to save listing in %s%s: %m', $filename,
				$_tmp_ext);
		return 0;
	};

	## Save each record
	$self->rewind();
	while (my $rec = $self->next()) {
		print $fh $rec->string(), "\n";
	}

	## Create and print end-of-file -1 string.
	my $rec = Tucows::Mirror::Listing::Record->new({blank => '-1'});
	print $fh $rec->string(), "\n";

	$fh->close() || do {
		logit('error', 3, 'E52 Unable to close %s%s: %m', $filename,
				$_tmp_ext);
		return 0;
	};

	rename($filename . $_tmp_ext, $filename) || do {
		logit('error', 3, 'E53 Unable to rename %s%s to %s: %m', $filename,
				$_tmp_ext, $filename);
		return 0;
	};

	return 1;
}

sub find {
	my $self = shift;
	my $name = shift;
	if (defined $self->{_lookup}->{$name}) {
		return $self->{_data}->[$self->{_lookup}->{$name}];
	} else {
		return undef;
	}
}

sub rewind {
	my $self = shift;
	$self->{_index} = 0;
}

sub next {
	my $self = shift;
	return $self->{_data}->[$self->{_index}++];
}

1;
