# TUCOPY Tucows mirroring software
# Copyright (C) 2001 Tucows Inc.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. 
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. 
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


#
# vi:set tabstop=4:
# $Id: Log.pm,v 1.17 2001/04/10 20:11:54 bjovanovic Exp $
#
# Logging facility
#
# Authors:
#	Leonid Igolnik <lim@tucows.com>
#	Bojan Jovanovic <bjovanovic@tucows.com>
#	Alex Nobert <anobert@tucows.com>
#
# 0 - internal priority can not be filtered
# 1 - Critical problems - system is unusable
# 2 - Severe errors
# 3 - Warnings
# 4 - Normal but significant condition
# 5 - Informational
# 6 and above - Debug


package Tucows::Mirror::Log;
use Tucows::Mirror::Config;
use IO::File;

#we export logit function by default
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(get_log_filenames logit set_verbose);

use strict;

#structure that holds all logging related details
my $_log_info = undef;
my $_verbose = 0;

#Set verbose - each msg will be sent to STDERR in adition to regular logs
sub set_verbose {
	my ($param) = @_;
	$_verbose = $param if defined $param;
	return $_verbose;
}

#Log msg to the log files
sub logit {
	my ($type, $level, @msg) = @_;
	my $caller = caller;

	#Init log data, if called for the first time
	_init_logs() unless $_log_info;

	#check current debug level, and return if msg should be ignored
	return if $level > $_log_info->{log_level};

	#Send msg to all the logs that are configured to accept messages of
	#the type same as current one
	foreach my $log (@{$_log_info->{types}->{$type}}) {

		my $bytes = _log_msg($log, $level, $caller, @msg);

		#Check if log reached max size, and rotate if necessary
		#Note this doesn't work well in multi-process environment
		$log->{_file_size} += $bytes;
		if ($log->{max_size} && $log->{_file_size} > $log->{max_size}) {
			_rotate_log($log);
		}
	}
}

#Rotate log
sub _rotate_log {
	my ($log) = @_;

	#close current file descriptor
	$log->{_fh}->close() || warn "Can't close log file: $log->{location}: $!\n";

	#move the file
	rename $log->{location}, $log->{location} . '.bak' ||
			warn "Can't rename $log->{location} to $log->{location}.bak: $!";
	
	#init new log
	_init_file($log);
}

#Massages the input msg, internal use only
sub _log_msg {
	my ($log, $level, $caller, @args) = @_;

	my $fh = $log->{_fh};

	#Get sprintf like format
	my $format = shift @args;

	#add caller name and pid
	if ($level > 0) {
		my @caller = split('::', $caller);
		$format = '%s[%pid] ' . $format;
		unshift @args, $caller[-1];
	}
	
	#Add timestamp
	$format = '%s ' . $format . "\n";
	my $time = time;
	# simplify parsing - have epoc as well
	unshift @args, '['.(scalar localtime($time))."] $time";

	#Replace all internal params
	$format =~ s/\%pid/$$/g;
	$format =~ s/\%m/$!/g;

	#Create a args and add the timestamp
	my $msg = sprintf($format, @args);

	#send to file
	$fh->print($msg);
	#send to screen if verbose is set
	print STDERR $msg if $_verbose;

	return length $msg;
}

#initialize all the file related things
sub _init_file {
	my ($log) = @_;

	#open the log file
	$log->{_fh} = IO::File->new($log->{location},
									O_WRONLY | O_APPEND | O_CREAT) || do {
			warn "Unable to initialize log file '$log->{location}': $!\n";
			return undef;
	};

	#Set autoflush, so the messages appear in the log immediately
	$log->{_fh}->autoflush(1);

	#record current file size, used for log rotation
	$log->{_file_size} = ($log->{_fh}->stat())[7];
	
	return 1;

}

#read log configuration and store it in internal structure
sub _init_logs {
	my $config = Tucows::Mirror::Config->new() || do {
		warn "Can't load default configuration file check that MirrorConfigFile is set\n";
		return undef;
	};
	my $params = $config->params();
	my $number_of_files = $config->log_max_number();
	$_log_info = {};
	$_log_info->{logs} = [];
	$_log_info->{types} = {};
	$_log_info->{log_level} = $config->{log_level};

	for(my $i = 0; $i <= $number_of_files; $i++) {
		my $log = $_log_info->{logs}->[$i] = {};
		$log->{location} = $params->{'log_location_' . $i};
		$log->{max_size} = $params->{'log_max_size_' . $i};
		$log->{types} = [split ',', $params->{'log_types_' . $i}];

		#Open file and initialize related values
		_init_file($log) || do {
			warn "Unable to initialize log file :", $log->{location};
			return undef;
		};
	
		#Data structure initialized below is a bit tricky, see notes below 
		#for more info	
		foreach my $type (@{$log->{types}}) {
			if (!defined $_log_info->{types}->{$type}) {
				$_log_info->{types}->{$type} = [];
			}
			push @{$_log_info->{types}->{$type}}, $log;
		}
	}

	#use Data::Dumper($_log_info->{logs}, $_log_info->{types}) to see how data
	#is stored in the _log_info;
	return 1;
}

sub get_log_filenames
	{
	my $log_type = shift;

	my @filenames;

	foreach my $destination (@{$_log_info->{types}->{$log_type}})
		{
		push @filenames, $destination->{location};
		}
	return \@filenames;
	}

1;
